{# Used for MP and NS trees #}
{% extends "admin/change_list.html" %}
{% load admin_list admin_tree i18n %}

{% block extrastyle %}
    {{ block.super }}
    {% treebeard_css %}
{% endblock %}

{% block extrahead %}
    {{ block.super }}
    {% treebeard_js %}
{% endblock %}

{% block result_list %}
    {% if action_form and actions_on_top and cl.full_result_count %}
        {% admin_actions %}
    {% endif %}
    {% result_tree cl request %}
    {% if action_form and actions_on_bottom and cl.full_result_count %}
        {% admin_actions %}
    {% endif %}
{% endblock %}
